/*
 * Copyright (C) 1998 Angus Mackay. All rights reserved.
 * See the file COPYING for details.
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifndef HAVE_SYS_IOCTL_H
#  error ioctl not found
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/soundcard.h>
#include <sys/ioctl.h>


#include "play.h"
#include "sinus.h"

/*
 * open the audio device for writing to
 */
int open_audio(char *dev, int bufsize, int bits, int rate, int channels)
{
  int fd = -1;
  int x;

  /*
   * Open the device driver:
   */
  if((fd=open(audio_dev,O_WRONLY)) < 0) 
  {
    printf("%s: Opening audio device %s\n",
        sys_errlist[errno],
        audio_dev);
    goto ERR;
  }

  /*
   * Determine the audio device's block size:
   */
  x = bufsize;
  if(ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &x) < 0)
  {
    fprintf(stderr, "%s: Optaining DSP's block size\n",sys_errlist[errno]);
    goto ERR;
  }

  /*
   * Check the range on the buffer sizes:
   */
  if (x < 4096 || x > 65536 ) {
    fprintf(stderr, "%s: Audio block size (%d bytes)\n",
        sys_errlist[errno=EINVAL], x);
    goto ERR;
  } 

  /*
   * Set the data bit size:
   */
  x = bits;
  if(ioctl(fd,SNDCTL_DSP_SAMPLESIZE,&x) < 0)
  {
    fprintf(stderr, "%s: Setting DSP to %u bits", sys_errlist[errno], bits);
    goto ERR;
  }
  if(x != bits)
  {
    fprintf(stderr, 
        "warning: the requested sample size was changed (%d -> %d)\n",
        bits, x);
    fprintf(stderr, "         bye the audio driver. It may not sound good.\n");
  }

  /*
   * Set the mode to be Stereo or Mono:
   */
  x = (channels == 2) ? 1 : 0;
  if(ioctl(fd,SNDCTL_DSP_STEREO,&x) < 0)
  {
    fprintf(stderr, "%s: Unable to set DSP to %s mode\n",
        sys_errlist[errno], (channels == 2) ? "Stereo":"Mono");
    goto ERR;
  }		

  /*
   * Set the sampling rate:
   */
  x = rate;
  if(ioctl(fd,SNDCTL_DSP_SPEED,&x) < 0)
  {
    fprintf(stderr, "%s: Unable to set audio sampling rate to %d\n",
        sys_errlist[errno], rate);
    goto ERR;
  }
  if(x != rate)
  {
    fprintf(stderr, 
        "warning: the requested sampling rate was changed (%d -> %d)\n",
        rate, x);
    fprintf(stderr, "         bye the audio driver. It may not sound good.\n");
  }

  return(fd);

ERR:
  if(fd >= 0) { close(fd); }
  return(-1);
}

/*
 * play the sample to the already opened file descriptor
 */
void play(int fd, void *data, int size, int bufsize)
{
  int i;
  char *p;
  int n;

  if(ioctl(fd,SNDCTL_DSP_SYNC,0) != 0)
  {
    fprintf(stderr, "%s: ioctl(%d,SNDCTL_DSP_SYNC,0)",
        sys_errlist[errno], fd);
    exit(1);
  }

  p = data;

  dprintf((stderr, "p: %p -> %p\n", p, &(p[size])));
  dprintf((stderr, "size: %06X\n", size));
  dprintf((stderr, "bufsize: %06X\n", bufsize));

  for(i=0; i<size; i+=bufsize)
  {
    n = (size - i < bufsize) ? (size - i) : bufsize;
    /* dprintf((stderr, "&(p[%06X]): %p, n: %X\n", i, &(p[i]), n)); */
    if(write(fd, &(p[i]), n) != n)
    {
      dprintf((stderr, "write on %d: %s\n", fd, strerror(errno)));
      fprintf(stderr, "write on %d: %s\n", fd, strerror(errno));
      exit(1);
    }
  }

  if(ioctl(fd,SNDCTL_DSP_SYNC,0) != 0)
  {
    fprintf(stderr, "%s: ioctl(%d,SNDCTL_DSP_SYNC,0)",
        sys_errlist[errno], fd);
    exit(1);
  }
}

