/*
 * sinus
 *
 * Copyright (C) 1998 Angus Mackay. All rights reserved.
 * See the file COPYING for details.
 *
 * A neat little piece of code that generates waves (and WAVs).
 * it will also play them!
 *
 * Angus Mackay
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_GETOPT_H
#  include <getopt.h>
#endif

#include "sized_types.h"
#include "byteorder.h"
#include "sinus.h"

#include "play.h"

#define WRITE_CHUNK 8192
#define DEF_AUDIO_DEV "/dev/dsp"

char *program_name;

char *audio_dev = NULL;
char *output_file = NULL;

int g_bits = 16;
int g_channels = 2;
int g_rate = 44100;

int signed_data = 1;

int options = 0;

/**************************************/

void print_useage( void )
{
  fprintf(stdout, "useage: ");
  fprintf(stdout, 
      "%s [options] <freq1[/freq1R][:freq2[/freq2R]][,duration]> ...\n\n", 
      program_name);
  fprintf(stdout, " Options are:\n");
  fprintf(stdout, "  -b, --bits <n>\tset output to n bit audio\n");
  fprintf(stdout, "  -c, --channels <n>\tset output to n channels (2 == Stereo)\n");
  fprintf(stdout, "  -d, --device <dev>\tsend audio to dev\n");
  fprintf(stdout, "  -n, --noplay \t\tdo not play the sample\n");
  fprintf(stdout, "  -o, --output <file>\twrite sample to file in MS WAV format\n");
  fprintf(stdout, "  -r, --rate <n>\tset sampling rate to n\n");
#ifdef DEBUG
  fprintf(stdout, "  -D, --debug\t\tturn on debuggin\n");
#endif
  fprintf(stdout, "  -H, --help\t\tdisplay this help and exit\n");
  fprintf(stdout, "  -V, --version\t\toutput version information and exit\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "expamples:\n");
  fprintf(stdout, "  - \"%s 400:100,0.5\" would create a 0.5 second wave form that\n", program_name);
  fprintf(stdout, "    started at 400 Hz and ended at 100 Hz\n");
  fprintf(stdout, "  - \"%s 100/200:200/300,1.5\" would create a 0.5 second wave form that\n", program_name);
  fprintf(stdout, "    started at 100 Hz in the left channel and 200 Hz in the right and\n");
  fprintf(stdout, "    ended at 200 Hz in the left channel and 300 Hz in the right\n");
  fprintf(stdout, "  - \"%s 600:1400,.2 1400,.5 1400:600,.2\" would create a wave form\n", program_name);
  fprintf(stdout, "    that started at 600 Hz, went to 1400 Hz, stayed ther for 0.5\n");
  fprintf(stdout, "    seconds, then return to 600 Hz\n");
  fprintf(stdout, "\n");
}

void print_version( void )
{
  fprintf(stdout, "%s: - v%s - %s\n", program_name, VERSION, "Angus Mackay");
}

#ifdef HAVE_GETOPT_LONG
#  define xgetopt(x1,x2,x3,x4,x5) getopt_long(x1,x2,x3,x4,x5)
#else
#  define xgetopt(x1,x2,x3,x4,x5) getopt(x1,x2,x3)
#endif

void parse_args(int argc, char **argv)
{
#ifdef HAVE_GETOPT_LONG
  struct option long_options[] = {
      {"bits",      required_argument, 0, 'b'},
      {"channels",  required_argument, 0, 'c'},
      {"device",    required_argument, 0, 'd'},
      {"noplay",    no_argument,       0, 'n'},
      {"output",    required_argument, 0, 'o'},
      {"rate",      required_argument, 0, 'r'},
      {"debug",     no_argument,       0, 'D'},
      {"help",      no_argument,       0, 'H'},
      {"version",   no_argument,       0, 'V'},
      {0,0,0,0}
  };
#else
#  define long_options NULL
#endif
  int opt;

  while((opt = xgetopt(argc, argv, "b:c:d:no:r:DHV", long_options, NULL)) != -1)
  {
    switch (opt)
    {
      case 'b':
        g_bits = atoi(optarg);
        dprintf((stderr, "g_bits: %d\n", g_bits));
        break;

      case 'c':
        g_channels = atoi(optarg);
        dprintf((stderr, "g_channels: %d\n", g_channels));
        break;

      case 'd':
        if(audio_dev)
        {
          free(audio_dev);
        }
        audio_dev = strdup(optarg);
        dprintf((stderr, "audio_dev: %s\n", audio_dev));
        break;

      case 'n':
        options |= OPT_NOPLAY;
        dprintf((stderr, "audio output suppressed\n"));
        break;

      case 'o':
        if(output_file)
        {
          free(output_file);
        }
        output_file = strdup(optarg);
        dprintf((stderr, "output_file: %s\n", output_file));
        break;

      case 'r':
        g_rate = atoi(optarg);
        dprintf((stderr, "g_rate: %d\n", g_rate));
        break;

      case 'D':
        options |= OPT_DEBUG;
        dprintf((stderr, "debugging on\n"));
        break;

      case 'H':
        print_useage();
        exit(0);
        break;

      case 'V':
        print_version();
        exit(0);
        break;

      default:
        fprintf(stderr, "Try `%s --help' for more information\n", argv[0]);
        exit(1);
        break;
    }
  }
}

/*
 * the function that actually generates the waveform.
 */
void build_wave(struct wave_t *wave, double *freq1, double *freq2)
{
  int cur_samp;
  int ind;
  double sin_val[2];
  double freq_step[2];
  double freq[2];
  double swing;
  double mean;
  int i;
  int bytes;

  swing = (wave->audio_max - wave->audio_min)/2;
  mean = (wave->audio_max + wave->audio_min)/2;
  dprintf((stderr, "swing: %.2f, mean: %.2f\n", swing, mean));

  bytes = wave->point_size;

  // why is the /2 there? I don't know enough about frequency response 
  // to say but it is neccessary
  // UPDATE: well after some searching I have found that it is because what
  //         I am really doing is taking the integral of the instantaneous
  //         frequency funtion (which is linear, y(t) = t + c) and thus getting
  //         y(s) = s(s/2 + c) which is where the 2 comes in.
  for(i=0; i<wave->channels; i++)
  {
    freq_step[i] = (freq2[i] - freq1[i])/2/(double)wave->samples;
  }

  for(i=0; i<wave->channels; i++)
  {
    freq[i] = freq1[i];
  }
  for(cur_samp=ind=0; cur_samp<wave->samples; cur_samp++)
  {
    dprintf((stderr, "sample: %8d", cur_samp));
    for(i=0; i<wave->channels; i++)
    {
      sin_val[i] = sin((freq[i])*((double)cur_samp/wave->rate)*2*M_PI);
      wave->pack((wave->data+ind), mean + (swing*sin_val[i]));
      dprint((stderr, ", c%d: %6.0f", i, wave->unpack(wave->data+ind)));
      dprint((stderr, ", freq[%d]: %7.2f", i, freq[i]));
      ind += bytes;
      freq[i] += freq_step[i];
    }
    dprint((stderr, "\n"));
  }
}

void wave_to_little_endian(struct wave_t *wave)
{
#ifdef __LITTLEENDIAN__
  return;
#else
  int i;
  void *p;

  switch(wave->point_size)
  {
    case 1:
      return;
    case 2:
      p = wave->data;
      for(i=0; i<wave->data_size; i+=2)
      {
        dprintf((stderr, "host -> le: %04X ->", (UINT16)*((UINT16 *)(p+i))));
        (INT16)*((UINT16 *)(p+i)) = HOST_TO_LENDIAN_INT16(*((UINT16 *)(p+i)));
        dprint((stderr, " %04X\n", (UINT16)*((UINT16 *)(p+i))));
      }
      return;
    case 4:
      p = wave->data;
      for(i=0; i<wave->data_size; i+=4)
      {
        dprintf((stderr, "host -> le: %08X ->", (UINT32)*((UINT32 *)(p+i))));
        (INT32)*((UINT32 *)(p+i)) = HOST_TO_LENDIAN_INT32(*((UINT32 *)(p+i)));
        dprint((stderr, " %08X\n", (UINT32)*((UINT32 *)(p+i))));
      }
      return;
    default:
      fprintf(stderr, "bad wave data (%d bits [%d] bytes)\n", wave->bits, wave->point_size);
      break;
  }
#endif
}

/*
 * write the PCM data to a MS RIFF WAV file
 */
int write_wav(char *filename, struct wave_t *wave)
{
  int fd;
  struct wavFormatChunk wf = { "fmt " };
  char *p;
  INT32 x;
  int i;
  int n;
  int wave_size;

  wave_size = wave->data_size;

  if((fd=open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
  { 
    fprintf(stderr, "error opening \"%s\" for writing: %s\n", filename, strerror(errno));
    goto ERR;
  }

  // start header
  if(write(fd, "RIFF    WAVE", 12) != 12)
  {
    fprintf(stderr, "write(\"%s\") failed: %s\n", filename, strerror(errno));
    goto ERR;
  }

  wf.chunkSize = HOST_TO_LENDIAN_INT32(sizeof(wf) - sizeof(wf.chunkID) - 
      sizeof(wf.chunkSize));
  wf.FormatTag = HOST_TO_LENDIAN_INT16((INT16)1); // PCM formatted audio
  wf.Channels = HOST_TO_LENDIAN_INT16((UINT16)(wave->channels));
  wf.SamplesPerSec = HOST_TO_LENDIAN_INT32((UINT32)(wave->rate));
  wf.AvgBytesPerSec = HOST_TO_LENDIAN_INT32(
      (UINT32)((wave->rate)*(wave->channels)*(wave->point_size)));
  wf.BlockAlign = HOST_TO_LENDIAN_INT16(
      (UINT16)((wave->channels)*(wave->point_size)));
  wf.BitsPerSample = HOST_TO_LENDIAN_INT16((UINT16)(wave->bits));
  write(fd, &wf, sizeof(wf));

  // start data
  if(write(fd, "data", 4) != 4)
  {
    fprintf(stderr, "write(\"%s\") failed: %s\n", filename, strerror(errno));
    goto ERR;
  }

  x = HOST_TO_LENDIAN_INT32(wave_size);
  // data size
  if(write(fd, &x, 4) != 4)
  {
    fprintf(stderr, "write(\"%s\") failed: %s\n", filename, strerror(errno));
    goto ERR;
  }

  // convert to little endian
  wave_to_little_endian(wave);

  // write the data
  p = (char *)(wave->data);
  for(i=0; i<wave_size; i+=WRITE_CHUNK)
  {
    n = (wave_size - i < WRITE_CHUNK) ? (wave_size - i) : WRITE_CHUNK;
    /* dprintf((stderr, "&(p[%06X]): %p, n: %X\n", i, &(p[i]), n)); */
    if(write(fd, &(p[i]), n) != n)
    {
      dprintf((stderr, "write on %d: %s\n", fd, strerror(errno)));
      fprintf(stderr, "write on %d: %s\n", fd, strerror(errno));
      goto ERR;
    }
  }

  close(fd);
  return 0;

ERR:
  if(fd >= 0) { close(fd); }
  return(-1);
}

/*
 * pack and unpack functions for handling 8,16,32 bit data
 * transparently
 */
void pack_uint8(void *p, double val) { *(UINT8 *)p = val; }
void pack_int16(void *p, double val) { *(INT16 *)p  = val; }
void pack_int32(void *p, double val) { *(INT32 *)p  = val; }

double unpack_uint8(void *p) { return(*(UINT8 *)p); }
double unpack_int16(void *p) { return(*(INT16 *)p);  }
double unpack_int32(void *p) { return(*(INT32 *)p);  }

/*
 * deceifer a command in the format:
 *   <f1[/f1R][:f2[/f2R]][,dur]>
 */
int parse_command(char *s, double *f1, double *f2, double *duration)
{
  char *t1;
  char *t2;
  char *t3 = NULL; // this MUST be initialized to NULL
  char *t4;

  if(s == NULL) 
  { 
    return(-1);
  }

  // find the important points in the string
  if((t4=strchr(s, ','))) { *t4++ = '\0'; }
  if((t2=strchr(s, ':'))) { *t2++ = '\0'; }
  if((t1=strchr(s, '/'))) { *t1++ = '\0'; }
  if(t2)
  {
    if((t3=strchr(t2, '/'))) { *t3++ = '\0'; }
  }
  dprintf((stderr, "s: %s, t1: %s, t2: %s, t3: %s, t4: %s\n", 
        s, t1 ? t1 : "(null)", t2 ? t2 : "(null)", 
        t3 ? t3 : "(null)", t4 ? t4 : "(null)"));

  f1[0] = atof(s);
  if(t1) { f1[1] = atof(t1); } else { f1[1] = f1[0]; }
  if(t2) { f2[0] = atof(t2); } else { f2[0] = f1[0]; }
  if(t3) { f2[1] = atof(t3); } else { f2[1] = f2[0]; }
  if(t4) { *duration = atof(t4); } else { *duration = 1; }

  return 0;
}

int main(int argc, char **argv)
{
  int fd;
  int bufsiz;
  double freq1[2];
  double freq2[2];
  double duration;
  double total_duration = 0;
  struct wave_t wave;
  struct wave_t w;
  int i;
  int j;

  // sanity check
  if(sizeof(INT8) != 1 || sizeof(INT16) != 2 || sizeof(INT32) != 4)
  {
    fprintf(stderr, "this program was not compilied properly!\n");
    exit(1);
  }

  program_name = argv[0];

  parse_args(argc, argv);

  // check options and set defaults 
  if(audio_dev == NULL) 
  {
    audio_dev = strdup(DEF_AUDIO_DEV);
  }
  switch(g_bits)
  {
    case 8:  
      signed_data = 0;
      wave.pack = w.pack = pack_uint8; 
      wave.unpack = w.unpack = unpack_uint8;
      break;
    case 16: 
      signed_data = 1; 
      wave.pack = w.pack = pack_int16; 
      wave.unpack = w.unpack = unpack_int16; 
      break;
    case 32: 
      signed_data = 1; 
      wave.pack = w.pack = pack_int32; 
      wave.unpack = w.unpack = unpack_int32; 
      break;
    default:
      fprintf(stderr, "only 8,16,32 bit audio is supported\n");
      exit(1);
      break;
  }
  if(g_channels < 1 || g_channels > 2)
  { 
    fprintf(stderr, "only Stereo and Mono modes are supported\n");
    exit(1);
  }
  if(output_file == NULL && (options & OPT_NOPLAY))
  { 
    fprintf(stderr, "nothing to do\n");
    exit(0);
  }
  if(optind == argc)
  {
    print_useage();
    exit(1);
  }

  wave.bits = g_bits;
  wave.rate = g_rate;
  wave.channels = g_channels;
  wave.samples = 0;
  wave.data_size = 0;
  wave.data = NULL;
  wave.point_size = ceil((float)wave.bits/8);

  // process the commands
  for(i=optind; i<argc; i++)
  {
    dprintf((stderr, "argv[%d]: %s\n", i, argv[i]));
    parse_command(argv[i], freq1, freq2, &duration);

    // check for user sillyness
    for(j=0; j<wave.channels; j++)
    {
      if(freq1[j] > wave.rate/2 || freq2[j] > wave.rate/2)
      {
        fprintf(stderr, 
            "warning: you have requested a frequency higher than 1/2 of your\n");
        fprintf(stderr, 
            "         sampling rate(%d), Nyquest says no to this but I will\n",
            wave.rate/2);
        fprintf(stderr, "         let it slide.\n");
      }
    }

    w.bits = wave.bits;
    w.rate = wave.rate;
    w.channels = wave.channels;
    w.samples = duration * w.rate;
    w.point_size = wave.point_size;
    w.data_size = w.samples*(w.point_size)*w.channels;
    if(signed_data)
    {
      w.audio_max = pow(2, w.bits-1) - 1;
      w.audio_min = -(pow(2, w.bits-1));;
    }
    else
    {
      w.audio_max = pow(2, w.bits) - 1;
      w.audio_min = 0;
    }

    dprintf((stderr, "audio_max: %.0f, audio_min: %.0f\n", 
          w.audio_max, w.audio_min));
    dprintf((stderr, "%d samples using %d bytes of data\n", 
          w.samples, w.data_size));

    // update the sizes
    wave.data_size += w.data_size;
    wave.samples += w.samples;
    total_duration += duration;

    // allocate memory for our wave form
    if((wave.data=realloc(wave.data, wave.data_size)) == NULL)
    {
      perror("out of memory");
      exit(1);
    }

    // point this segment in to the real wave
    w.data = (wave.data+wave.data_size-w.data_size);

    dprintf((stderr, "wave.data: %p, w.data: %p\n", wave.data, w.data));

    // build our wave form
    if(w.channels == 2)
    {
      printf("building %d sample (%.2f s, %.1f kB) wave %.1f/%.1f -> %.1f/%.1f\n", 
          w.samples, duration, (float)w.data_size/1024, 
          freq1[0], freq1[1], freq2[0], freq2[1]);
    }
    else
    {
      printf("building %d sample (%.2f s, %.1f kB) wave %.1f -> %.1f\n", 
          w.samples, duration, (float)w.data_size/1024, freq1[0], freq2[0]);
    }
    build_wave(&w, freq1, freq2);
  }

  printf("built a %d sample (%.2f s, %.1f kB) wave\n", 
      wave.samples, total_duration, (float)wave.data_size/1024);

  // play the sample
  if(!(options & OPT_NOPLAY))
  {
    printf("playing %.2f s %d Hz %d bit %s wave to \"%s\"\n", 
        total_duration, wave.rate, wave.bits, 
        wave.channels == 2 ? "Stereo" : "Mono", audio_dev);
    if((fd=open_audio(audio_dev, bufsiz, wave.bits, wave.rate, 
            wave.channels)) == -1)
    {
      fprintf(stderr, "failed to open the audio device\n");
    }
    else
    {
      play(fd, wave.data, wave.data_size, bufsiz);
      close(fd);
    }
  }

  // write the wave to a .wav file
  if(output_file != NULL)
  {
    printf("writing %.2f kB sample to \"%s\"\n", 
        (float)(sizeof(struct wavFormatChunk) + wave.data_size)/1024, output_file);
    write_wav(output_file, &wave);
  }

  free(wave.data);

  return 0;
}
